#include <UIPEthernet.h>
#include <EEPROM.h>
#include <OneWire.h>
#include <DallasTemperature.h>

#define SWSERIAL
// #define DEBUG
// #define DEBUG_SERIAL // Debugging der Seriellenkommunikation zum KeyPad via Serielle Schnittstelle
// #define DEBUG_NET // Debugging der Netzwerkkommunikation über die Serielle Schnittstelle
// #define NO_KEY_PAD // Abschalten des LOCK durch Fehlen des Key Pad
// #define CLEAR_CODE // Setzt Code Zähler beim Booten auf 0, Neuer Code muss eingegeben werden
// #define CLEAR_IP // Setzt die IP auf den Vorgabewert zurück

#ifdef SWSERIAL
    #include <SoftwareSerial.h>
    
    SoftwareSerial SSerial(2,3); // RX, TX
#endif

// door Keller
// #define MACADDRESS 0x80,0x02,0x54,0x1F,0xAB,0xED
// #define MYIPADDR 192,168,234,161
// door carport 
// #define MACADDRESS 0x80,0x00,0x54,0x9F,0xAC,0xEA
// #define MYIPADDR 192,168,234,165
// door devel 
#define MACADDRESS 0x80,0x01,0x43,0x1A,0xCD,0xEB
#define MYIPADDR 192,168,234,166
// door EG
// #define MACADDRESS 0x90,0xA2,0xDA,0x30,0x86,0x11 
// #define MYIPADDR 192,168,234,167
// door-garage
// #define MACADDRESS 0x80,0xB3,0x12,0x9C,0x4C,0x21
// #define MYIPADDR 192,168,234,168

// Port Config
#define LISTENPORT 2340

#define RELAY 5

#define TIME_OPEN 10000

#define AutoReSync 10000

struct baseConf
{
    uint8_t ip[4];
    long time_open;
    byte min_wait;
    byte max_wait;
};
baseConf MyConf;

EthernetServer server = EthernetServer(LISTENPORT);
EthernetClient client;
EthernetClient newClient;
uint8_t mac[6] = {MACADDRESS};
uint8_t ip[4] = {MYIPADDR};

const byte MaxCodeLength = 12;
const byte MaxCodeNumber = 16;
byte NrOfCodes = 0;
char codes[MaxCodeNumber][MaxCodeLength];

const long presharedKey = 13071978;

const unsigned long CommuncationKey = 611311;
char KeyPadBuffer[MaxCodeLength];
byte BufferCount = 0;

union key {
  byte as_array[4];
  long as_long;
} convert;

#define MIN_WAIT 5
byte wait = MIN_WAIT;
long time_now = 0;
unsigned int timer = 0;
byte status = 0; // 0 Fehlerfrei, 1 Tech. Error Lock, 2 Manual Lock, 3 Key Error Lock

#define ONEwire 8
#define DS18B20_RES 12
DeviceAddress DS18B20_ADR;
OneWire oneWire(ONEwire);
DallasTemperature DS18B20(&oneWire);

#define ClearEEProm A5
#define ClearIP A2
#define reSyncPIN A0

void SerWrite(byte val)
{
    #ifdef SWSERIAL
        SSerial.write(val);
    #elif
        Serial1.write(val);
    #endif
}

void setup()
{
    Serial.begin(9600);
    Serial.println(F("DG1ATN Zahlenschloss Server"));
    #ifdef SWSERIAL
        SSerial.begin(9600);
    #elif
        Serial1.begin(9600);
    #endif

    // mir gerade nicht klar wofür der ist
    // digitalWrite(7,HIGH);

    delay(10);

    SerWrite(0);

    reSync();


    // EEPROM lesen
    EEPROM.get(0,MyConf);
    #ifdef DEBUG
        Serial.print(F("MyConf: "));
        Serial.println(int(sizeof(MyConf)));
        Serial.print(F("IP: "));
        if (MyConf.ip[0] == 255)
        {
            Serial.println(MyConf.ip[0]);
            Serial.println(MyConf.ip[1]);
            Serial.println(MyConf.ip[2]);
            Serial.println(MyConf.ip[3]);
        }        
        else
            Serial.println(MyConf.ip[0]);

    #endif
    byte e = int(sizeof(MyConf));
    NrOfCodes = EEPROM.read(e);

    #ifdef DEBUG
        Serial.print(F("NrOfCodes:"));
        Serial.println(NrOfCodes);
    #endif

    #ifdef CLEAR_CODE
        NrOfCodes = 0;
        Serial.println(F("Clear Codes"));
    #else
        pinMode(ClearEEProm,INPUT_PULLUP);
        if (digitalRead(ClearEEProm) == LOW)
        {
            NrOfCodes = 0;
            Serial.println(F("Clear Codes"));
        }
    #endif

    if (NrOfCodes == 255)
        NrOfCodes = 0;
    for (byte d = 0; d < NrOfCodes+1; d++)
    {
        for (byte i = 0; i < MaxCodeLength; i++)
        {
            codes[d][i] = EEPROM.read(d * MaxCodeLength + e + 1 +i);
        }
        #ifdef DEBUG
            Serial.print(F("Code["));
            Serial.print(d);
            Serial.print(F("]: "));
            Serial.println(codes[d]);
        #endif
    }

    // SS Pin beim every by default: D8
    Ethernet.init(10);
    // SS Pin beim nano by default: D10
    # ifdef CLEAR_IP
        MyConf.ip[0] = 255;
        Serial.println(F("Clear IP"));
    #else
        pinMode(ClearIP,INPUT_PULLUP);
        if (digitalRead(ClearIP) == LOW)
        {
            MyConf.ip[0] = 255;
            Serial.println(F("Clear IP"));
        }
    #endif
    if (MyConf.ip[0] == 255) 
    {
        MyConf.ip[0] = ip[0]; 
        MyConf.ip[1] = ip[1]; 
        MyConf.ip[2] = ip[2]; 
        MyConf.ip[3] = ip[3]; 
    }
    if (MyConf.min_wait == 255)
        MyConf.min_wait = MIN_WAIT;
    convert.as_long = MyConf.time_open;
    if (convert.as_array[0] == 255)
        MyConf.time_open = TIME_OPEN;
    Ethernet.begin(mac, MyConf.ip);
    server.begin();
    Serial.println(Ethernet.localIP());  

    // Relay Ausgang vorbereiten
    pinMode(RELAY,OUTPUT);
    digitalWrite(RELAY, LOW);

    // ReSync Eingang vorbereiten
    pinMode(reSyncPIN,INPUT_PULLUP);


    // Fehlermeldung
    if (NrOfCodes == 0)
        SerWrite(130);

    // Temeperatursensor einrichten
    DS18B20.begin();
    Serial.print(F("Sesoren: "));
    Serial.println(DS18B20.getDeviceCount(), DEC);
    if (DS18B20.getAddress(DS18B20_ADR,1))
        DS18B20.setResolution(DS18B20_ADR,DS18B20_RES);

    DS18B20.requestTemperaturesByAddress(DS18B20_ADR);
    
    Serial.println(DS18B20.getTempC(DS18B20_ADR));

    // Kommunikation am Ende der Boot Sequenz aufbauen
    // 10 s delay, damit der Arduion an der Tastatur auch fertig booten kann 
    delay(10000);
    SerWrite(250);
    // Auslesen der seriellen Verbindung
    #ifdef SWSERIAL
        while (SSerial.available())
        { 
            byte comKey = SSerial.read();
        }
    #elif
        while ((Serial1.available())
        { 
            byte comKey = Serial1.read();
        }
    #endif

    delay(1000);
    reSync();

}

bool checkKey(char Key[MaxCodeLength], byte length)
{
    #ifdef DEBUG_SERIAL
        Serial.println(F("checkKey: "));
        Serial.print(Key);
        Serial.print(F(" / "));
        Serial.print(length);
        Serial.print(F(" / "));
        Serial.println(MaxCodeNumber);
    #endif
    
    if (length < 1) 
        return(true);
    bool ErrKey = true;
    for (byte i=0; i < MaxCodeNumber; i++)
    {
        bool KeyOK = true;
        byte j = 0;
        while ((KeyOK) && (j < length) && (Key[j] != '\0'))
        {
            if (Key[j] == codes[i][j])
                KeyOK = true;
            else
                KeyOK = false;
            j++;
        }  
        if (KeyOK)
        {
            #ifdef DEBUG_SERIAL
                Serial.print(F("codes[i]: "));
                Serial.println(int(codes[i][length+1]));
                Serial.print(F("sizeof(codes[i])/length: "));
                Serial.print(sizeof(codes[i]));
                Serial.print(F(" / "));
                Serial.println(length);
                Serial.print(F("code[i+1]: "));
                Serial.println(int(codes[i+1][0]));
            #endif
            if ((int(codes[i][length+1])==0) || (sizeof(codes[i])==length) ||
                (int(codes[i][length+1])==int(codes[i+1][0])))
            {
                Serial.println(F("OK"));
                ErrKey = false;
            }
        }                    
    }
    return(ErrKey);
}

bool openDoor(char Key[MaxCodeLength], byte length)
{
    bool ErrKey = checkKey(Key,length);
    if (not ErrKey)
    {
            SerWrite(120);
            digitalWrite(RELAY,HIGH);
            startTimer(MyConf.time_open);
            wait = MyConf.min_wait;
            ErrKey = false;
    }
    return(ErrKey);
}

void restBuffer()
{
    for (byte i=0; i <= BufferCount; i++)
        KeyPadBuffer[i] = 0;
    BufferCount = 0;
    SerWrite(200);
}

void startTimer(unsigned int t)
{
    timer = t;
    time_now = millis();
}

void stopTimer()
{
    timer = 0;
}

void execTimer()
{
    if (timer > 0)
    {
        if (millis() - time_now > timer) 
        {
            if (status == 1)
            {
                reSync();
            }
            else if (digitalRead(RELAY))
            {
                digitalWrite(RELAY,LOW);            
                SerWrite(200);
            }
            else 
            {
                if (wait < MyConf.max_wait)
                    wait = 2 * wait;
                SerWrite(200);
            }
                status = 0;
                stopTimer();
        }
    }
}

void reSync()
{
    SerWrite(150);

    convert.as_long = random(CommuncationKey);
    byte RecInititalKeyLength = 0;
    while (RecInititalKeyLength < 4)
    {
        #ifdef DEBUG
            // Serial.println(convert.as_array[RecInititalKeyLength]);
        #endif        
        #ifdef SWSERIAL
            SSerial.write(convert.as_array[RecInititalKeyLength]);
        #elif
            Serial1.write(convert.as_array[RecInititalKeyLength]);
        #endif
        RecInititalKeyLength++;
    }
    #ifdef DEBUG
        Serial.print(F("New Key: "));
        Serial.println(convert.as_long);
    #endif
    randomSeed(convert.as_long+presharedKey); 
    status = 0;           
}

void loop()
{ 
    execTimer();

    byte decodKey = 0;
    #ifdef SWSERIAL
        if (SSerial.available())
        { 
            byte comKey = SSerial.read();
    #elif
        if (Serial1.available())
        { 
            byte comKey = Serial1.read();
    #endif
    #ifdef DEBUG_SERIAL
        Serial.print(F("comKey: "));
        Serial.println(comKey);
    #endif
        byte randNumber = random(0, 254);
        decodKey = comKey ^ randNumber;      
    #ifdef DEBUG_SERIAL
        Serial.print(F("decodKey: "));
        Serial.println(decodKey);
    #endif
        if (char(decodKey) == '#')
        {
            #ifdef DEBUG_SERIAL
                Serial.println(KeyPadBuffer);
                Serial.println(BufferCount);
            #endif
            bool ErrKey = openDoor(KeyPadBuffer,BufferCount);
            if (ErrKey)
            {
                Serial.println(F("KEY Wrong"));
                SerWrite(130);
                status = 3;
                startTimer(wait*1000);
            }
                
            KeyPadBuffer[BufferCount] = char(decodKey);
            BufferCount ++  ;
        }
        else
        if (char(decodKey) == '*')
        {
            restBuffer();
            Serial.println(F("Clear"));
            SerWrite(200);
        }
        else
        if ((decodKey > 47) & (decodKey < 69))
        {
            if (KeyPadBuffer[BufferCount-1] == '#')
                restBuffer();
            KeyPadBuffer[BufferCount] = char(decodKey);
            BufferCount ++;
            byte a = BufferCount % 2;
            SerWrite(101+a);
        }
        else
        {
            #ifdef NO_KEY_PAD
                restBuffer();
                status = 9;
            #else
// Lock per Ethernet deaktivieren, Key neu übertragen
                Serial.println(F("Lock"));
// Serial.println(decodKey);
                SerWrite(250);
                status = 1;
                startTimer(AutoReSync);
            #endif
        }

    }
    if (digitalRead(reSyncPIN) == LOW)
    {
        if (status == 1)
        {
            reSync();
            Serial.println(F("reSync Manual"));
        }
    }
    if (client = server.available())
    {
        byte i = 0;
        char ethCommand[MaxCodeLength + 15] = "";
        while (client.available() > 0)
        {
            char in = client.read();
            if (in == '\n')
            {
                ethCommand[i] = '\0';
                break;
            }
            else
            {
                ethCommand[i] = in;
                i++;
            }
        }

        String s(ethCommand);
        s.toUpperCase();

        #ifdef DEBUG_NET
            Serial.println(s);
        #endif

        bool t = false;
        if (s.startsWith(F("READ_KEY:")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                client.print(F("LAST KEY: "));
                client.println(KeyPadBuffer);
            }
        }
        if (s.startsWith(F("NEW_CODE:")))
        {
    	    if (NrOfCodes > MaxCodeNumber)
            {
                client.println(F("ERROR MAX CODE"));
            }
            else
            {            
                bool ok = false;
                s.remove(0,9);
                char code[MaxCodeLength];
                #ifdef DEBUG_NET
                    Serial.print(F("NrOfCodes:"));
                    Serial.println(NrOfCodes);
                #endif
                if (NrOfCodes == 0)
                {    
                    ok = true;
                } 
                else
                {
                    s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
                    ok = not checkKey(code,s.indexOf(F(":")));
                    s.remove(0,s.indexOf(F(":"))+1);         
                }
                #ifdef DEBUG_NET
                    Serial.print(F("new code ok:"));
                    Serial.println(ok);
                #endif
                if (ok)
                {
                    byte e = int(sizeof(MyConf));
                    s.toCharArray(code,MaxCodeLength);
                    #ifdef DEBUG
                        Serial.print(F("SIZE MyConf:"));
                        Serial.println(e);
                        Serial.print(F("Code: "));
                        Serial.println(s);
                    #endif

                    if (s.indexOf(F(":")) == -1)
                    {
                        for (byte i = 0; i<sizeof(code); i++)
                            codes[NrOfCodes][i] = code[i];
                        for (byte i = 0; i<sizeof(code); i++)
                        {
                            EEPROM.update(NrOfCodes*MaxCodeLength + e + 1 + i,code[i]);
                        }
                        NrOfCodes ++;
                        EEPROM.update(e,NrOfCodes);
                        client.println(F("CODE OK"));
                    }
                    else
                        client.println(F("CODE NOT VALID"));
                }
                else
                    client.println(F("UNAUTHORIZED"));
            }
            if (NrOfCodes > 0)
                SerWrite(200);
        }
        if (s.startsWith(F("READ_CODE:")))
        {
            // Nur mit gültigem Key auslesen
            s.remove(0,10);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            #ifdef DEBUG_NET
                Serial.print(F("code: "));
                Serial.println(code);
                Serial.print(F("sizeof(code): "));
                Serial.println(sizeof(code));
            #endif
            if (not checkKey(code,sizeof(code)))
            {
                for (byte i = 0; i < NrOfCodes; i++)
                {
                    client.print(i);
                    client.print(":");
                    client.println(codes[i]);
                }
            }
            else
                client.println(F("UNAUTHORIZED"));
        }
        if (s.startsWith(F("DEL_CODE:")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
            if (not checkKey(code,s.indexOf(F(":"))))
            {
                byte e = int(sizeof(MyConf));
                s.remove(0,s.indexOf(F(":"))+1);         
                for (byte d = s.toInt(); d < NrOfCodes - 1; d++)
                {
                    for (byte i = 0; i < sizeof(codes[d+1]); i++)
                    {
                        codes[d][i] = codes[d+1][i];
                        EEPROM.update(e + d * MaxCodeLength + 1 + i,codes[d][i]);
                    }
                }
                NrOfCodes --;
                EEPROM.update(e,NrOfCodes);
                client.println(F("DELETE OK"));
            }
            else
                client.println(F("UNAUTHORIZED"));
        }
        if (s.startsWith(F("OPEN_DOOR:")))
        {
            s.remove(0,10);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = openDoor(code,sizeof(code));
            if (ErrKey)
            {
                client.println(F("KEY Wrong"));
                SerWrite(130);
                startTimer(wait*1000);
            }
            else
            {
                client.println(F("OK"));
            }
                
            // reset buffer
            restBuffer();
        }
        if (s.startsWith(F("RST_WAIT:")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                wait = MyConf.min_wait;
                stopTimer();
                restBuffer();
                SerWrite(200);
                client.println(F("OK"));
            }
        }
        if (s.startsWith(F("READ_WAIT:")))
        {
            s.remove(0,10);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                client.print(F("WAIT: "));
                client.println(wait);
            }
        }
        if (s.startsWith(F("TEMP_IN")))
        {
            DS18B20.requestTemperaturesByAddress(DS18B20_ADR);
            
            float temp = DS18B20.getTempC(DS18B20_ADR);
// Serial.println(temp);
            client.print(F("Temp IN: "));
            client.println(temp);            
        }
        if (s.startsWith(F("RE_SYNC")))
        {
            s.remove(0,8);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
                reSync();
        }
        if (s.startsWith(F("TEMP_OUT")))
        {
            SerWrite(50);
            t = true;
        }
        if (s.startsWith(F("READ_LIGHT")))
        {
            SerWrite(55);
            t = true;
        }
        if (s.startsWith(F("LIGHT_ON")))
        {
            SerWrite(56);
            t = true;
        }
        if (s.startsWith(F("LIGHT_OFF")))
        {
            SerWrite(57);
            t = true;
        }
        if (s.startsWith(F("SET_IP")))
        {
            s.remove(0,7);
            char code[MaxCodeLength];
            s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,s.indexOf(F(":")));
            s.remove(0,s.indexOf(F(":"))+1);         
            if (not ErrKey)
            {
                for (byte i = 0; i < 4; i++)
                {
                    MyConf.ip[i]=s.substring(0,s.indexOf(F("."))).toInt();
                    s.remove(0,s.indexOf(F("."))+1);
                }
                EEPROM.put(0,MyConf);
                client.println(F("DONE"));
            }
        }
        if (s.startsWith(F("SET_OPEN")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,s.indexOf(F(":")));
            s.remove(0,s.indexOf(F(":"))+1);         
            #ifdef DEBUG
                Serial.print(F("SET_OPEN: "));
                Serial.print(code);
                Serial.print(" > ");
                Serial.println(ErrKey);
            #endif
            if (not ErrKey)
            {
                MyConf.time_open=s.toInt();
                EEPROM.put(0,MyConf);
                #ifdef DEBUG
                    Serial.print(F(" "));
                    Serial.println(MyConf.time_open);
                #endif
                client.println(F("DONE"));
            }
        }
        if (s.startsWith(F("GET_OPEN")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                client.print(F("Open Time: "));
                client.println(MyConf.time_open);
            }
        }
        if (s.startsWith(F("SET_MIN_WAIT")))
        {
            s.remove(0,13);
            char code[MaxCodeLength];
            s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,s.indexOf(F(":")));
            s.remove(0,s.indexOf(F(":"))+1);         
            if (not ErrKey)
            {
                MyConf.min_wait=s.toInt();
                EEPROM.put(0,MyConf);
                client.println(F("DONE"));
            }
        }
        if (s.startsWith(F("GET_MIN_WAIT")))
        {
            s.remove(0,13);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                client.print(F("Min Wait: "));
                client.println(MyConf.min_wait);
            }
        }
        if (s.startsWith(F("SET_MAX_WAIT")))
        {
            s.remove(0,13);
            char code[MaxCodeLength];
            s.substring(0,s.indexOf(F(":"))).toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,s.indexOf(F(":")));
            s.remove(0,s.indexOf(F(":"))+1);         
            if (not ErrKey)
            {
                MyConf.max_wait=s.toInt();
                EEPROM.put(0,MyConf);
                client.println(F("DONE"));
            }
        }
        if (s.startsWith(F("GET_MAX_WAIT")))
        {
            s.remove(0,13);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                client.print(F("Max Wait: "));
                client.println(MyConf.max_wait);
            }
        }
        if (s.startsWith(F("SET_DARK")))
        {
            s.remove(0,9);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                SerWrite(65);
                t = true;
            }
        }
        if (s.startsWith(F("GET_DARK")))
        {
            SerWrite(66);
            t = true;
        }
        if (s.startsWith(F("SET_LIGHT")))
        {
            s.remove(0,10);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
            {
                SerWrite(60);
                t = true;
            }
        }
        if (s.startsWith(F("GET_LIGHT")))
        {
            SerWrite(61);
            t = true;
        }
        if (s.startsWith(F("READ_STATUS")))
        {
            client.print(F("Status: "));
            client.println(status);
        }
        if (s.startsWith(F("LOCK_DOOR")))
        {
            s.remove(0,10);
            char code[MaxCodeLength];
            s.toCharArray(code,MaxCodeLength);
            bool ErrKey = checkKey(code,sizeof(code));
            if (not ErrKey)
                SerWrite(250);
                status = 2;
        }

        // Antwort vom Tastenfeld
        while (t)
        {
            #ifdef SWSERIAL
            if (SSerial.available())
            {
                char c = SSerial.read();
            #elif
            if (Serial1.available())
            {
                char c = Serial1.read();
            #endif
                client.print(c);
                if (c == 10)
                {
                    client.println();
                    t = false;
                }
            }
        }
    }    
    if (NrOfCodes == 0)
        SerWrite(130);
}